var gl;
var currentProgram;

function initGL(canvas) {
   try {
      gl = canvas.getContext("experimental-webgl");
      gl.viewportWidth = canvas.width;
      gl.viewportHeight = canvas.height;
   } catch (e) {
      alert("Could not initialize WebGL, sorry :(");
   }
   if (!gl) {
      alert("Could not initialize WebGL, sorry :(");
   }
}

function getShader(gl, filePath, isFragment) {
   var shader;
   var shaderSource;

   if (isFragment) {
      shader = gl.createShader(gl.FRAGMENT_SHADER);
   } else {
      shader = gl.createShader(gl.VERTEX_SHADER);
   }

   $.ajax({
      url: filePath,
      type: 'GET',
      dataType: "text",
      async: false,
      success: function(res) {
         shaderSource = res;
      },
      error: function(jqXHR, textStatus, errorThrown) {
         alert("AJAX Error");
         alert(textStatus);
         alert(errorThrown);
      }
   });

   gl.shaderSource(shader, shaderSource);
   gl.compileShader(shader);

   if (! gl.getShaderParameter(shader, gl.COMPILE_STATUS)) {
      alert(gl.getShaderInfoLog(shader));
      return null;
   }

   return shader;
}

function createProgram(fragmentShaderID, vertexShaderID) {
   var fragmentShader = getShader(gl, fragmentShaderID, true);
   var vertexShader = getShader(gl, vertexShaderID, false);

   var program = gl.createProgram();
   gl.attachShader(program, vertexShader);
   gl.attachShader(program, fragmentShader);
   gl.linkProgram(program);

   if (!gl.getProgramParameter(program, gl.LINK_STATUS)) {
      alert("Could not initialize shaders");
   }

   program.vertexPositionAttribute = gl.getAttribLocation(program, "aVertexPosition");
   gl.enableVertexAttribArray(program.vertexPositionAttribute);

   program.vertexNormalAttribute = gl.getAttribLocation(program, "aVertexNormal");
   gl.enableVertexAttribArray(program.vertexNormalAttribute);

   program.textureCoordAttribute = gl.getAttribLocation(program, "aTextureCoord");
   gl.enableVertexAttribArray(program.textureCoordAttribute);

   program.pMatrixUniform = gl.getUniformLocation(program, "uPMatrix");
   program.mvMatrixUniform = gl.getUniformLocation(program, "uMVMatrix");
   program.nMatrixUniform = gl.getUniformLocation(program, "uNMatrix");
   program.samplerUniform = gl.getUniformLocation(program, "uSampler");
   program.useTexturesUniform = gl.getUniformLocation(program, "uUseTextures");
   program.useLightingUniform = gl.getUniformLocation(program, "uUseLighting");
   program.ambientColorUniform = gl.getUniformLocation(program, "uAmbientColor");
   program.pointLightingLocationUniform = gl.getUniformLocation(program, "uPointLightingLocation");
   program.pointLightingColorUniform = gl.getUniformLocation(program, "uPointLightingColor");
   program.useCisternUniform = gl.getUniformLocation(program, "uCistern");
   program.useWaterHeightUniform = gl.getUniformLocation(program, "uWaterHeight");

   return program;
}

function initShaders() {
   currentProgram = createProgram("Shaders/fragment.glsl", "Shaders/vertex.glsl");
}

var maltaMesh;
var maltaWater;

var pins = [];
var cisterns = [];

var mvMatrix = mat4.create();
var mvMatrixStack = [];
var pMatrix = mat4.create();

function degToRad(degrees) {
   return degrees * Math.PI / 180;
}

var xRot = 0;
var xSpeed = 0;

var yRot = 0;
var ySpeed = 0;

var zVal = -.8;

var filter = 0;
var currentlyPressedKeys = {};

function handleKeyDown(event) {
   currentlyPressedKeys[event.keyCode] = true;

   if (String.fromCharCode(event.keyCode) == "F") {
      filter += 1;
      if (filter == 3) {
         filter = 0;
      }
   }
}

function handleKeyUp(event) {
   currentlyPressedKeys[event.keyCode] = false;
}

function handleKeys() {
   if (currentlyPressedKeys[33]) {
      // Page Up
      zVal -= 0.05;
   }
   if (currentlyPressedKeys[34]) {
      // Page Down
      zVal += 0.05;
   }
   if (currentlyPressedKeys[37]) {
      // Left cursor key
      yRot -= 1;
   }
   if (currentlyPressedKeys[39]) {
      // Right cursor key
      yRot += 1;
   }
   if (currentlyPressedKeys[38]) {
      // Up cursor key
      xRot -= 1;
   }
   if (currentlyPressedKeys[40]) {
      // Down cursor key
      xRot += 1;
   }
}

function loadWorld() {
   loadCisternFiles();

   // Malta vertex file
   var requestMalta = new XMLHttpRequest();
   requestMalta.open("GET", "TopoData/Malta-fractalD.txt", false);
   requestMalta.send();
   maltaMesh = new Mesh(requestMalta.responseText, "Textures/Stones.jpg",
                        true, false, new Vertex(999, 999, 999),
                        new Vertex(-999, -999, -999), 0);

   // Malta water
   maltaWaterMesh = new Mesh(null, "Textures/waterTexture.jpg", true, true,
                             maltaMesh.min, maltaMesh.max, 0);
   loadPins();
}

function loadPins() {
   var requestCube = new XMLHttpRequest();
   requestCube.open("GET", "Vertices/Cube.txt", false);
   requestCube.send();
   pins[0] = new Mesh(requestCube.responseText, "Textures/cube.png", false, false,
                      new Vertex(999, 999, 999), new Vertex(-999, -999, -999), 0.0);

   requestCube = new XMLHttpRequest();
   requestCube.open("GET", "Vertices/Cube.txt", false);
   requestCube.send();
   pins[1] = new Mesh(requestCube.responseText, "Textures/cubeBlue.png", false, false,
                      new Vertex(999, 999, 999), new Vertex(-999, -999, -999), 0.0);

   requestCube = new XMLHttpRequest();
   requestCube.open("GET", "Vertices/Cube.txt", false);
   requestCube.send();
   pins[2] = new Mesh(requestCube.responseText, "Textures/cubePurple.png", false, false,
                      new Vertex(999, 999, 999), new Vertex(-999, -999, -999), 0.0);

}

var lastVal = -1;

function drawScene() {
   gl.viewport(0, 0, gl.viewportWidth, gl.viewportHeight);
   gl.clear(gl.COLOR_BUFFER_BIT | gl.DEPTH_BUFFER_BIT);

   mat4.perspective(45, gl.viewportWidth / gl.viewportHeight, 0.1, 100.0, pMatrix);
   gl.useProgram(currentProgram);

   // Ambient lighting
   var ambientR = 0.2;
   var ambientG = 0.2;
   var ambientB = 0.2;

   // Point lighting
   var pointR = 0.8;
   var pointG = 0.8;
   var pointB = 0.8;

   // Light Position
   var lightPosX = 0;
   var lightPosY = 0;
   var lightPosZ = 0;

   var lighting = document.getElementById("lighting").checked;
   gl.uniform1i(currentProgram.useLightingUniform, lighting);

   if (lighting) {
      gl.uniform3f(
         currentProgram.ambientColorUniform,
         parseFloat(ambientR),
         parseFloat(ambientG),
         parseFloat(ambientB)
      );

      gl.uniform3f(
         currentProgram.pointLightingLocationUniform,
         parseFloat(lightPosX),
         parseFloat(lightPosY),
         parseFloat(lightPosZ)
      );

      gl.uniform3f(
         currentProgram.pointLightingColorUniform,
         parseFloat(pointR),
         parseFloat(pointG),
         parseFloat(pointB)
      );
   }

   var textures = document.getElementById("textures").checked;
   gl.uniform1i(currentProgram.useTexturesUniform, textures);

   gl.uniform1f(currentProgram.useWaterHeightUniform, maltaWaterMesh.waterHeight);

   mat4.identity(mvMatrix);

   var current = document.getElementById("cisterns").selectedIndex;

   if (document.getElementById("cisterns").value == "MaltaMap") {
      gl.uniform1i(currentProgram.useCisternUniform, false);

      if (current != lastVal) {
         zVal = -0.8;
         xRot = 0.0;
         yRot = 0.0;
         lastVal = current;
         console.log("TEST");
      }

      mvPushMatrix();
         mat4.translate(mvMatrix, [-(maltaMesh.max.x + maltaMesh.min.x)/2,
                        -(maltaMesh.max.y + maltaMesh.min.y)/2, zVal]);
         mat4.rotate(mvMatrix, degToRad(xRot), [1, 0, 0]);
         mat4.rotate(mvMatrix, degToRad(yRot), [0, 1, 0]);

         mvPushMatrix();
            maltaMesh.Draw();
            if (document.getElementById("water").checked)
               maltaWaterMesh.Draw();
         mvPopMatrix();

         gl.uniform1i(currentProgram.useCisternUniform, true);

         for (var i = 0; i < maltaMesh.pinVertices.length; i++) {
            mvPushMatrix();
               mat4.translate(mvMatrix, [maltaMesh.pinVertices[i].x - 14.6,
                              maltaMesh.pinVertices[i].y - 35.6, (maltaMesh.pinVertices[i].z)]);
               mat4.scale(mvMatrix, [0.005, 0.005, 0.005]);

               pins[i].Draw();
            mvPopMatrix();
         }
      mvPopMatrix();
   }
   else {
      gl.uniform1i(currentProgram.useCisternUniform, true);

      if (current != lastVal) {
         zVal = -7.0;
         xRot = 0.0;
         yRot = 0.0;
         lastVal = current;
      }
      mvPushMatrix();
         mat4.translate(mvMatrix, [-(cisterns[current].rock.max.x + cisterns[current].rock.min.x)/2,
                        -(cisterns[current].rock.max.y + cisterns[current].rock.min.y)/2, zVal]);
         mat4.rotate(mvMatrix, degToRad(xRot), [1, 0, 0]);
         mat4.rotate(mvMatrix, degToRad(yRot), [0, 1, 0]);

         cisterns[current].drawCistern(document.getElementById("water").checked);
      mvPopMatrix();
   }
}

function tick() {
   requestAnimFrame(tick);
   handleKeys();
   drawScene();
}

function webGLStart() {
   var canvas = document.getElementById("canvas");
   initGL(canvas);
   initShaders();
   loadWorld();

   gl.clearColor(1.0, 1.0, 1.0, 1.0);
   gl.enable(gl.DEPTH_TEST);

   document.onkeydown = handleKeyDown;
   document.onkeyup = handleKeyUp;

   tick();
}

function loadCisternFiles() {
   var gps = [
              new Vertex(0, 0, 0),
              new Vertex(-0.1, 0.2, 0.1),
              new Vertex(-0.2, 0.1, 0),
              new Vertex(0, 0, 0),
              new Vertex(0.1, -0.1, 0),
              new Vertex(0.2, -0.2, 0.1)
             ];


   var names = [
               "M'dina Cathedral Sacristy",
               "House Dar Ta'anna",
               "Convento Dei Cappuccinni",
               "Case Cutietta",
               "Site 8",
               "Tas Silg"
               ];

   var numCisterns = document.getElementById("cisterns").length - 1;
   currentProgram.cistern = true;

   for (var i = 0; i < numCisterns; i++) {
      cisterns[i] = new Cistern(names[i], gps[i], "Vertices/" +
                                document.getElementById("cisterns").options[i].value);
      cisterns[i].loadCistern();
   }
}
