function Mesh(data, textureFile, malta, water, min, max, zSubtraction) {
   this.stride = 0;
   this.min = min;
   this.max = max;
   this.zSubtraction = zSubtraction;

   this.data = data;

   if (malta && !water)
      this.buffers = this.ScanMalta();
   else if (malta)
      this.buffers = this.ScanMaltaWater();
   else
      this.buffers = this.ScanData();

   this.textureFile = textureFile;
   this.texture = this.InitTexture();
}

// Regular data file
Mesh.prototype.ScanData = function() {
   var lines = this.data.split("\n");
   var vertexCount = 0;
   var index = 0;
   var textureIndex = 0;

   // Positions
   var vertexPositions = [];
   var vertexTextureCoords = [];
   var vertexNormals = [];

   var vertices = [];
   var textures = [];

   var v1;
   var v2;
   var vector;

   for (var i in lines) {
      var vals = lines[i].replace(/^\s+/, "").split(/\s+/);
      if (vals.length == 4 && vals[0] == "v") {
         // Scan vertices
         vertices[index] = new Vertex(parseFloat(vals[1]), parseFloat(vals[2]),
                                      parseFloat(vals[3]));
         this.CalculateExtremes(vertices, index);

         index += 1;
      }
      else if (vals.length == 3 && vals[0] == "vt") {
         // Scan texture coords
         textures[textureIndex] = new Texture(parseFloat(vals[1]), parseFloat(vals[2]));
         textureIndex += 1;
      }
      else if (vals.length == 4 && vals[0] == "f") {
         // Map vertices to each other
         v1 = vec3.create([vertices[vals[2]-1].x-vertices[vals[1]-1].x,
            vertices[vals[2]-1].y-vertices[vals[1]-1].y,
            vertices[vals[2]-1].z-vertices[vals[1]-1].z]);
         v2 = vec3.create([vertices[vals[3]-1].x-vertices[vals[1]-1].x,
            vertices[vals[3]-1].y-vertices[vals[1]-1].y,
            vertices[vals[3]-1].z-vertices[vals[1]-1].z]);

         v1 = vec3.normalize(v1);
         v2 = vec3.normalize(v2);

         vector = vec3.cross(v1, v2);

         for (var i = 0; i < 3; i++) {
            vertexNormals.push(vector[0]);
            vertexNormals.push(vector[1]);
            vertexNormals.push(vector[2]);
         }

         for (var i = 0; i < 3; i++) {
            vertexPositions.push(vertices[vals[i+1]-1].x);
            vertexPositions.push(vertices[vals[i+1]-1].y);
            vertexPositions.push(vertices[vals[i+1]-1].z - this.zSubtraction);
         }

         for (var i = 0; i < 3; i++) {
            vertexTextureCoords.push(textures[vals[i+1]-1].x);
            vertexTextureCoords.push(textures[vals[i+1]-1].y);
         }

         vertexCount += 3;
      }
   }

   return this.BindGLBuffers(vertexCount, vertexPositions, vertexTextureCoords,
                             vertexNormals);
}

// Malta
Mesh.prototype.ScanMalta = function () {
   var vertices = this.ManageData();

   var vertexCount = 0;
   var maltaVertexPositions = [];
   var maltaVertexTextureCoords = [];
   var maltaVertexNormals = [];

   for (var i = 0; i < this.stride - 1; i++) {
      for (var j = 0; j < this.stride - 1; j++) {
         var index = j + i*this.stride;

         // Calculate vertex positions
         maltaVertexPositions.push(vertices[index].x);
         maltaVertexPositions.push(vertices[index].y);
         maltaVertexPositions.push(vertices[index].z);

         maltaVertexPositions.push(vertices[index + 1].x);
         maltaVertexPositions.push(vertices[index + 1].y);
         maltaVertexPositions.push(vertices[index + 1].z);

         maltaVertexPositions.push(vertices[index + this.stride].x);
         maltaVertexPositions.push(vertices[index + this.stride].y);
         maltaVertexPositions.push(vertices[index + this.stride].z);

         maltaVertexPositions.push(vertices[index + 1].x);
         maltaVertexPositions.push(vertices[index + 1].y);
         maltaVertexPositions.push(vertices[index + 1].z);

         maltaVertexPositions.push(vertices[index + 1 + this.stride].x);
         maltaVertexPositions.push(vertices[index + 1 + this.stride].y);
         maltaVertexPositions.push(vertices[index + 1 + this.stride].z);

         maltaVertexPositions.push(vertices[index + this.stride].x);
         maltaVertexPositions.push(vertices[index + this.stride].y);
         maltaVertexPositions.push(vertices[index + this.stride].z);

         for (var k = 0; k < 2; k++) {
            // Calculate normals
            v1 = vec3.create([vertices[index].x-vertices[index + 1].x,
               vertices[index].y-vertices[index + 1].y,
               vertices[index].z-vertices[index + 1].z]);
            v2 = vec3.create([vertices[index + this.stride].x-vertices[index + 1].x,
               vertices[index + this.stride].y-vertices[index + 1].y,
               vertices[index + this.stride].z-vertices[index + 1].z]);

            v1 = vec3.normalize(v1);
            v2 = vec3.normalize(v2);

            var vector = vec3.cross(v2, v1);

            for (var l = 0; l < 3; l++) {
               maltaVertexNormals.push(parseFloat(vector[0]));
               maltaVertexNormals.push(parseFloat(vector[1]));
               maltaVertexNormals.push(parseFloat(vector[2]));
            }
         }

         // Calculate texture coords
         maltaVertexTextureCoords.push(parseFloat(j/this.stride));
         maltaVertexTextureCoords.push(parseFloat(i/this.stride));

         maltaVertexTextureCoords.push(parseFloat((j + 1)/this.stride));
         maltaVertexTextureCoords.push(parseFloat(i/this.stride));

         maltaVertexTextureCoords.push(parseFloat(j/this.stride));
         maltaVertexTextureCoords.push(parseFloat((i + 1)/this.stride));

         maltaVertexTextureCoords.push(parseFloat((j + 1)/this.stride));
         maltaVertexTextureCoords.push(parseFloat(i/this.stride));

         maltaVertexTextureCoords.push(parseFloat((j + 1)/this.stride));
         maltaVertexTextureCoords.push(parseFloat((i + 1)/this.stride));

         maltaVertexTextureCoords.push(parseFloat(j/this.stride));
         maltaVertexTextureCoords.push(parseFloat((i + 1)/this.stride));

         vertexCount += 6;
      }
   }

   this.CalculatePins(vertices);

   return this.BindGLBuffers(vertexCount, maltaVertexPositions, maltaVertexTextureCoords, maltaVertexNormals);
}

Mesh.prototype.CalculatePins = function(vertices) {
   this.pinVertices = [];

   var request = new XMLHttpRequest();
   request.open("GET", "Vertices/Pins.txt", false);
   request.send();

   var lines = request.responseText.split("\n");

   for (var i in lines) {
      var vals = lines[i].replace(/^\s+/, "").split(/\s+/);
      if (vals.length == 2)
         this.pinVertices[i] = new Vertex(parseFloat(vals[0]), parseFloat(vals[1]), 0.0);
   }

   var deltaY = (this.max.x - this.min.x)/this.stride;
   var deltaX = (this.max.y - this.min.y)/this.stride;

   for (var i = 0; i < this.pinVertices.length; i++) {
      var jPin = Math.floor((this.pinVertices[i].x - 14.6 - this.min.x)/deltaX);
      var iPin = Math.floor((this.pinVertices[i].y - 35.6 - this.min.y)/deltaY);

      this.pinVertices[i].z = vertices[(256 - iPin) * (this.stride) + jPin].z;

      /*console.log("vert x: " + vertices[(256 - iPin) * (this.stride) + jPin].x + " vert y: " + vertices[(256 - iPin) * (this.stride) + jPin].y);
      console.log("pin x: " + (this.pinVertices[i].x - 14.6) + " pin y: " + (this.pinVertices[i].y - 35.6));
      console.log("ipin: " + iPin + " jpin " + jPin);
      console.log((iPin * this.stride + jPin));
      console.log("z: " + ((this.pinVertices[i].z*(this.max.z - this.min.z)*7)+this.min.z));
      console.log("stride: " + this.stride);*/
   }
}

// Malta water
Mesh.prototype.ScanMaltaWater = function() {
   var height = ((0-this.min.z)/(this.max.z - this.min.z)/7) + 0.0017;
   this.waterHeight = height;

   var vertices = [
      this.min.x, this.max.y, height,
      this.max.x, this.min.y, height,
      this.min.x, this.min.y, height,
      this.min.x, this.max.y, height,
      this.max.x, this.max.y, height,
      this.max.x, this.min.y, height
   ];

   var textures = [
      0, 1,
      1, 0,
      0, 0,
      0, 1,
      1, 1,
      1, 0
   ];

   var normals = [
      0, 0, 1,
      0, 0, 1,
      0, 0, 1,
      0, 0, 1,
      0, 0, 1,
      0, 0, 1
   ];

   return this.BindGLBuffers(6, vertices, textures, normals);
}

Mesh.prototype.CalculateExtremes = function(vertices, i) {
   if (vertices[i].x < this.min.x)
      this.min.x = vertices[i].x;
   if (vertices[i].x > this.max.x)
      this.max.x = vertices[i].x;

   if (vertices[i].y < this.min.y)
      this.min.y = vertices[i].y;
   if (vertices[i].y > this.max.y)
      this.max.y = vertices[i].y;

   if (vertices[i].z < this.min.z)
      this.min.z = vertices[i].z;
   if (vertices[i].z > this.max.z)
      this.max.z = vertices[i].z;
}

Mesh.prototype.InitTexture = function() {
   var texture = gl.createTexture();
   texture.image = new Image();
   texture.image.onload = function () {
      HandleLoadedTexture(texture);
   }
   texture.image.src = this.textureFile;

   return texture;
}

Mesh.prototype.ManageData = function() {
   var lines = this.data.split("\n");
   var lastY = 0;
   var count = 0;
   var vertices = [];

   for (var i = 0; i < lines.length; ++ i) {
      var vals = lines[i].replace(/^\s+/, "").split(/\s+/);

      if (vals.length == 3 || (vals.length == 4 && vals[3] == "")) {
         vertices[i] = new Vertex(parseFloat(vals[0]) - 14.6, parseFloat(vals[1]) - 35.6, parseFloat(vals[2]));

         this.CalculateExtremes(vertices, i);

         // Set the stride (grid size)
         if (lastY == 0)
            lastY = vertices[i].y;
         else if (lastY != vertices[i].y && this.stride == 0)
            this.stride = i;

         count++;
      }
   }

   for (var i = 0; i < count; i++) {
      vertices[i].z = (vertices[i].z - this.min.z)/(this.max.z - this.min.z)/7;
   }

   return vertices;
}

Mesh.prototype.BindGLBuffers = function(count, vertexPositions, vertexTextureCoords, vertexNormals) {
   var vertexPositionBuffer, vertexTextureCoordBuffer, vertexNormalBuffer;

   vertexPositionBuffer = gl.createBuffer();
   gl.bindBuffer(gl.ARRAY_BUFFER, vertexPositionBuffer);
   gl.bufferData(gl.ARRAY_BUFFER, new Float32Array(vertexPositions), gl.STATIC_DRAW);
   vertexPositionBuffer.itemSize = 3;
   vertexPositionBuffer.numItems = count;

   vertexTextureCoordBuffer = gl.createBuffer();
   gl.bindBuffer(gl.ARRAY_BUFFER, vertexTextureCoordBuffer);
   gl.bufferData(gl.ARRAY_BUFFER, new Float32Array(vertexTextureCoords), gl.STATIC_DRAW);
   vertexTextureCoordBuffer.itemSize = 2;
   vertexTextureCoordBuffer.numItems = count;

   vertexNormalBuffer = gl.createBuffer();
   gl.bindBuffer(gl.ARRAY_BUFFER, vertexNormalBuffer);
   gl.bufferData(gl.ARRAY_BUFFER, new Float32Array(vertexNormals), gl.STATIC_DRAW);
   vertexNormalBuffer.itemSize = 3;
   vertexNormalBuffer.numItems = count;

   return new Buffers(vertexPositionBuffer, vertexTextureCoordBuffer, vertexNormalBuffer, null);
}

Mesh.prototype.Draw = function() {
   gl.bindBuffer(gl.ARRAY_BUFFER, this.buffers.positionBuffer);
   gl.vertexAttribPointer(currentProgram.vertexPositionAttribute, this.buffers.positionBuffer.itemSize, gl.FLOAT, false, 0, 0);

   gl.bindBuffer(gl.ARRAY_BUFFER, this.buffers.normalBuffer);
   gl.vertexAttribPointer(currentProgram.vertexNormalAttribute, this.buffers.normalBuffer.itemSize, gl.FLOAT, false, 0, 0);

   gl.bindBuffer(gl.ARRAY_BUFFER, this.buffers.textureCoordBuffer);
   gl.vertexAttribPointer(currentProgram.textureCoordAttribute, this.buffers.textureCoordBuffer.itemSize, gl.FLOAT, false, 0, 0);

   gl.activeTexture(gl.TEXTURE0);
   gl.bindTexture(gl.TEXTURE_2D, this.texture);
   gl.uniform1i(currentProgram.samplerUniform, 0);

   setMatrixUniforms();
   gl.drawArrays(gl.TRIANGLES, 0, this.buffers.positionBuffer.numItems);
}

// Global functions
function mvPushMatrix() {
   var copy = mat4.create();
   mat4.set(mvMatrix, copy);
   mvMatrixStack.push(copy);
}

function mvPopMatrix() {
   if (mvMatrixStack.length == 0) {
      throw "Invalid popMatrix!";
   }
   mvMatrix = mvMatrixStack.pop();
}

function HandleLoadedTexture(texture) {
   gl.pixelStorei(gl.UNPACK_FLIP_Y_WEBGL, true);
   gl.bindTexture(gl.TEXTURE_2D, texture);
   gl.texImage2D(gl.TEXTURE_2D, 0, gl.RGBA, gl.RGBA, gl.UNSIGNED_BYTE, texture.image);
   gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MAG_FILTER, gl.LINEAR);
   gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MIN_FILTER, gl.LINEAR_MIPMAP_NEAREST);
   gl.generateMipmap(gl.TEXTURE_2D);

   gl.bindTexture(gl.TEXTURE_2D, null);
}

function setMatrixUniforms() {
   gl.uniformMatrix4fv(currentProgram.pMatrixUniform, false, pMatrix);
   gl.uniformMatrix4fv(currentProgram.mvMatrixUniform, false, mvMatrix);

   var normalMatrix = mat3.create();
   mat4.toInverseMat3(mvMatrix, normalMatrix);
   mat3.transpose(normalMatrix);
   gl.uniformMatrix3fv(currentProgram.nMatrixUniform, false, normalMatrix);
}
